pub struct TimeSeries { /* private fields */ }Expand description
An iterator of a sequence of evenly spaced Epochs.
(Python documentation hints) :type start: Epoch :type end: Epoch :type step: Duration :type inclusive: bool
Implementations
impl TimeSeries
impl TimeSeries
pub fn exclusive(start: Epoch, end: Epoch, step: Duration) -> TimeSeries ⓘ
pub fn exclusive(start: Epoch, end: Epoch, step: Duration) -> TimeSeries ⓘ
Return an iterator of evenly spaced Epochs, inclusive on start and exclusive on end.
use hifitime::{Epoch, Unit, TimeSeries};
let start = Epoch::from_gregorian_utc_at_midnight(2017, 1, 14);
let end = Epoch::from_gregorian_utc_at_noon(2017, 1, 14);
let step = Unit::Hour * 2;
let time_series = TimeSeries::exclusive(start, end, step);
let mut cnt = 0;
for epoch in time_series {
println!("{}", epoch);
cnt += 1
}
assert_eq!(cnt, 6)pub fn first_epoch(&self) -> Epoch
pub fn first_epoch(&self) -> Epoch
Returns first Epoch of this TimeSeries, without consuming the iterator.
pub fn last_epoch(&self) -> Epoch
pub fn last_epoch(&self) -> Epoch
Returns last Epoch of this TimeSeries, without consuming the iterator.
pub fn inclusive(start: Epoch, end: Epoch, step: Duration) -> TimeSeries ⓘ
pub fn inclusive(start: Epoch, end: Epoch, step: Duration) -> TimeSeries ⓘ
Return an iterator of evenly spaced Epochs, inclusive on start and on end.
use hifitime::{Epoch, Unit, TimeSeries};
let start = Epoch::from_gregorian_utc_at_midnight(2017, 1, 14);
let end = Epoch::from_gregorian_utc_at_noon(2017, 1, 14);
let step = Unit::Hour * 2;
let time_series = TimeSeries::inclusive(start, end, step);
let mut cnt = 0;
for epoch in time_series {
println!("{}", epoch);
cnt += 1
}
assert_eq!(cnt, 7)Examples found in repository?
35fn main() -> Result<(), Box<dyn Error>> {
36 pel::init();
37
38 // ====================== //
39 // === ALMANAC SET UP === //
40 // ====================== //
41
42 // Dynamics models require planetary constants and ephemerides to be defined.
43 // Let's start by grabbing those by using ANISE's MetaAlmanac.
44
45 let output_folder: PathBuf = [env!("CARGO_MANIFEST_DIR"), "../data", "04_output"]
46 .iter()
47 .collect();
48
49 let data_folder: PathBuf = [env!("CARGO_MANIFEST_DIR"), "examples", "04_lro_od"]
50 .iter()
51 .collect();
52
53 let meta = data_folder.join("lro-dynamics.dhall");
54
55 // Load this ephem in the general Almanac we're using for this analysis.
56 let mut almanac = MetaAlmanac::new(meta.to_string_lossy().as_ref())
57 .map_err(Box::new)?
58 .process(true)
59 .map_err(Box::new)?;
60
61 let mut moon_pc = almanac.get_planetary_data_from_id(MOON).unwrap();
62 moon_pc.mu_km3_s2 = 4902.74987;
63 almanac.set_planetary_data_from_id(MOON, moon_pc).unwrap();
64
65 let mut earth = almanac.get_planetary_data_from_id(EARTH).unwrap();
66 earth.mu_km3_s2 = 398600.436;
67 almanac.set_planetary_data_from_id(EARTH, earth).unwrap();
68
69 // Save this new kernel for reuse.
70 // In an operational context, this would be part of the "Lock" process, and should not change throughout the mission.
71 almanac
72 .planetary_data
73 .values()
74 .next()
75 .unwrap()
76 .save_as(&data_folder.join("lro-specific.pca"), true)?;
77
78 // Lock the almanac (an Arc is a read only structure).
79 let almanac = Arc::new(almanac);
80
81 // Orbit determination requires a Trajectory structure, which can be saved as parquet file.
82 // In our case, the trajectory comes from the BSP file, so we need to build a Trajectory from the almanac directly.
83 // To query the Almanac, we need to build the LRO frame in the J2000 orientation in our case.
84 // Inspecting the LRO BSP in the ANISE GUI shows us that NASA has assigned ID -85 to LRO.
85 let lro_frame = Frame::from_ephem_j2000(-85);
86
87 // To build the trajectory we need to provide a spacecraft template.
88 let sc_template = Spacecraft::builder()
89 .mass(Mass::from_dry_and_prop_masses(1018.0, 900.0)) // Launch masses
90 .srp(SRPData {
91 // SRP configuration is arbitrary, but we will be estimating it anyway.
92 area_m2: 3.9 * 2.7,
93 coeff_reflectivity: 0.96,
94 })
95 .orbit(Orbit::zero(MOON_J2000)) // Setting a zero orbit here because it's just a template
96 .build();
97 // Now we can build the trajectory from the BSP file.
98 // We'll arbitrarily set the tracking arc to 24 hours with a five second time step.
99 let traj_as_flown = Traj::from_bsp(
100 lro_frame,
101 MOON_J2000,
102 &almanac,
103 sc_template,
104 5.seconds(),
105 Some(Epoch::from_str("2024-01-01 00:00:00 UTC")?),
106 Some(Epoch::from_str("2024-01-02 00:00:00 UTC")?),
107 Aberration::LT,
108 Some("LRO".to_string()),
109 )?;
110
111 println!("{traj_as_flown}");
112
113 // ====================== //
114 // === MODEL MATCHING === //
115 // ====================== //
116
117 // Set up the spacecraft dynamics.
118
119 // Specify that the orbital dynamics must account for the graviational pull of the Earth and the Sun.
120 // The gravity of the Moon will also be accounted for since the spaceraft in a lunar orbit.
121 let mut orbital_dyn = OrbitalDynamics::point_masses(vec![EARTH, SUN, JUPITER_BARYCENTER]);
122
123 // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
124 // We're using the GRAIL JGGRX model.
125 let mut jggrx_meta = MetaFile {
126 uri: "http://public-data.nyxspace.com/nyx/models/Luna_jggrx_1500e_sha.tab.gz".to_string(),
127 crc32: Some(0x6bcacda8), // Specifying the CRC32 avoids redownloading it if it's cached.
128 };
129 // And let's download it if we don't have it yet.
130 jggrx_meta.process(true)?;
131
132 // Build the spherical harmonics.
133 // The harmonics must be computed in the body fixed frame.
134 // We're using the long term prediction of the Moon principal axes frame.
135 let moon_pa_frame = MOON_PA_FRAME.with_orient(31008);
136 let sph_harmonics = GravityField::new(GravityFieldData::from_shadr(
137 &jggrx_meta.uri,
138 80,
139 80,
140 almanac.frame_info(moon_pa_frame)?,
141 )?);
142
143 // Include the spherical harmonics into the orbital dynamics.
144 orbital_dyn.accel_models.push(sph_harmonics);
145
146 // We define the solar radiation pressure, using the default solar flux and accounting only
147 // for the eclipsing caused by the Earth and Moon.
148 // Note that by default, enabling the SolarPressure model will also enable the estimation of the coefficient of reflectivity.
149 let srp_dyn = SolarPressure::new(vec![EARTH_J2000, MOON_J2000], &almanac)?;
150
151 // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
152 // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
153 let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
154
155 println!("{dynamics}");
156
157 // Now we can build the propagator.
158 let setup = Propagator::default_dp78(dynamics.clone());
159
160 // For reference, let's build the trajectory with Nyx's models from that LRO state.
161 let (sim_final, traj_as_sim) = setup
162 .with(*traj_as_flown.first(), almanac.clone())
163 .until_epoch_with_traj(traj_as_flown.last().epoch())?;
164
165 println!("SIM INIT: {:x}", traj_as_flown.first());
166 println!("SIM FINAL: {sim_final:x}");
167 // Compute RIC difference between SIM and LRO ephem
168 let sim_lro_delta = sim_final
169 .orbit
170 .ric_difference(&traj_as_flown.last().orbit)?;
171 println!("{traj_as_sim}");
172 println!(
173 "SIM v LRO - RIC Position (m): {:.3}",
174 sim_lro_delta.radius_km * 1e3
175 );
176 println!(
177 "SIM v LRO - RIC Velocity (m/s): {:.3}",
178 sim_lro_delta.velocity_km_s * 1e3
179 );
180
181 traj_as_sim.ric_diff_to_parquet(
182 &traj_as_flown,
183 output_folder.join("./04_lro_sim_truth_error.parquet"),
184 ExportCfg::default(),
185 )?;
186
187 // ==================== //
188 // === OD SIMULATOR === //
189 // ==================== //
190
191 // After quite some time trying to exactly match the model, we still end up with an oscillatory difference on the order of 150 meters between the propagated state
192 // and the truth LRO state.
193
194 // Therefore, we will actually run an estimation from a dispersed LRO state.
195 // The sc_seed is the true LRO state from the BSP.
196 let sc_seed = *traj_as_flown.first();
197
198 // Load the Deep Space Network ground stations.
199 // Nyx allows you to build these at runtime but it's pretty static so we can just load them from YAML.
200 let ground_station_file: PathBuf = [
201 env!("CARGO_MANIFEST_DIR"),
202 "examples",
203 "04_lro_od",
204 "dsn-network.yaml",
205 ]
206 .iter()
207 .collect();
208
209 let devices = GroundStation::load_named(ground_station_file)?;
210
211 let mut proc_devices = devices.clone();
212
213 // Increase the noise in the devices to accept more measurements.
214 for gs in proc_devices.values_mut() {
215 if let Some(noise) = &mut gs
216 .stochastic_noises
217 .as_mut()
218 .unwrap()
219 .get_mut(&MeasurementType::Range)
220 {
221 *noise.white_noise.as_mut().unwrap() *= 3.0;
222 }
223 }
224
225 // Typical OD software requires that you specify your own tracking schedule or you'll have overlapping measurements.
226 // Nyx can build a tracking schedule for you based on the first station with access.
227 let trkconfg_yaml: PathBuf = [
228 env!("CARGO_MANIFEST_DIR"),
229 "examples",
230 "04_lro_od",
231 "tracking-cfg.yaml",
232 ]
233 .iter()
234 .collect();
235
236 let configs: BTreeMap<String, TrkConfig> = TrkConfig::load_named(trkconfg_yaml)?;
237
238 // Build the tracking arc simulation to generate a "standard measurement".
239 let mut trk = TrackingArcSim::<Spacecraft, GroundStation>::with_seed(
240 devices.clone(),
241 traj_as_flown.clone(),
242 configs,
243 123, // Set a seed for reproducibility
244 )?;
245
246 trk.build_schedule(&almanac)?;
247 let arc = trk.generate_measurements(&almanac)?;
248 // Save the simulated tracking data
249 arc.to_parquet_simple(output_folder.join("04_lro_simulated_tracking.parquet"))?;
250
251 // We'll note that in our case, we have continuous coverage of LRO when the vehicle is not behind the Moon.
252 println!("{arc}");
253
254 // Now that we have simulated measurements, we'll run the orbit determination.
255
256 // ===================== //
257 // === OD ESTIMATION === //
258 // ===================== //
259
260 let sc = SpacecraftUncertainty::builder()
261 .nominal(sc_seed)
262 .frame(LocalFrame::RIC)
263 .x_km(0.5)
264 .y_km(0.5)
265 .z_km(0.5)
266 .vx_km_s(5e-3)
267 .vy_km_s(5e-3)
268 .vz_km_s(5e-3)
269 .build();
270
271 // Build the filter initial estimate, which we will reuse in the filter.
272 let mut initial_estimate = sc.to_estimate()?;
273 initial_estimate.covar *= 3.0;
274
275 println!("== FILTER STATE ==\n{sc_seed:x}\n{initial_estimate}");
276
277 // Build the SNC in the Moon J2000 frame, specified as a velocity noise over time.
278 let process_noise = ProcessNoise3D::from_velocity_km_s(
279 &[1e-12, 1e-12, 1e-12],
280 1 * Unit::Hour,
281 10 * Unit::Minute,
282 None,
283 );
284
285 println!("{process_noise}");
286
287 // We'll set up the OD process to reject measurements whose residuals are move than 3 sigmas away from what we expect.
288 let odp = SpacecraftKalmanOD::new(
289 setup,
290 KalmanVariant::ReferenceUpdate,
291 Some(SigmaRejection::default()),
292 proc_devices,
293 almanac.clone(),
294 )
295 .with_process_noise(process_noise);
296
297 let od_sol = odp.process_arc(initial_estimate, &arc)?;
298
299 let final_est = od_sol.estimates.last().unwrap();
300
301 println!("{final_est}");
302
303 let ric_err = traj_as_flown
304 .at(final_est.epoch())?
305 .orbit
306 .ric_difference(&final_est.orbital_state())?;
307 println!("== RIC at end ==");
308 println!("RIC Position (m): {:.3}", ric_err.radius_km * 1e3);
309 println!("RIC Velocity (m/s): {:.3}", ric_err.velocity_km_s * 1e3);
310
311 println!(
312 "Num residuals rejected: #{}",
313 od_sol.rejected_residuals().len()
314 );
315 println!(
316 "Percentage within +/-3: {}",
317 od_sol.residual_ratio_within_threshold(3.0).unwrap()
318 );
319 println!("Ratios normal? {}", od_sol.is_normal(None).unwrap());
320
321 od_sol.to_parquet(
322 output_folder.join("04_lro_od_results.parquet"),
323 ExportCfg::default(),
324 )?;
325
326 // Create the ephemeris
327 let ephem = od_sol.to_ephemeris("LRO rebuilt".to_string());
328 let ephem_start = ephem.start_epoch().unwrap();
329 let ephem_end = ephem.end_epoch().unwrap();
330 // Check that the covariance is PSD throughout the ephemeris by interpolating it.
331 for epoch in TimeSeries::inclusive(ephem_start, ephem_end, Unit::Minute * 5) {
332 ephem
333 .covar_at(
334 epoch,
335 anise::ephemerides::ephemeris::LocalFrame::RIC,
336 &almanac,
337 )
338 .unwrap_or_else(|e| panic!("covar not PSD at {epoch}: {e}"));
339 }
340 // Export as BSP!
341 ephem
342 .write_spice_bsp(
343 -85,
344 output_folder.join("04_lro_rebuilt.bsp").to_str().unwrap(),
345 None,
346 )
347 .expect("could not built BSP");
348 let new_almanac = Almanac::default()
349 .load(output_folder.join("04_lro_rebuilt.bsp").to_str().unwrap())
350 .unwrap();
351 new_almanac.describe(None, None, None, None, None, None, None, None);
352 let (spk_start, spk_end) = new_almanac.spk_domain(-85).unwrap();
353
354 assert!((ephem_start - spk_start).abs() < Unit::Microsecond * 1);
355 assert!((ephem_end - spk_end).abs() < Unit::Microsecond * 1);
356
357 // In our case, we have the truth trajectory from NASA.
358 // So we can compute the RIC state difference between the real LRO ephem and what we've just estimated.
359 // Export the OD trajectory first.
360 let od_trajectory = od_sol.to_traj()?;
361 // Build the RIC difference.
362 od_trajectory.ric_diff_to_parquet(
363 &traj_as_flown,
364 output_folder.join("04_lro_od_truth_error.parquet"),
365 ExportCfg::default(),
366 )?;
367
368 Ok(())
369}Trait Implementations
impl Clone for TimeSeries
impl Clone for TimeSeries
fn clone(&self) -> TimeSeries ⓘ
fn clone(&self) -> TimeSeries ⓘ
1.0.0 (const: unstable) · Sourcefn clone_from(&mut self, source: &Self)
fn clone_from(&mut self, source: &Self)
source. Read moreimpl Debug for TimeSeries
impl Debug for TimeSeries
impl DerefToPyAny for TimeSeries
impl Display for TimeSeries
impl Display for TimeSeries
impl DoubleEndedIterator for TimeSeries
impl DoubleEndedIterator for TimeSeries
fn next_back(&mut self) -> Option<Epoch>
fn next_back(&mut self) -> Option<Epoch>
Sourcefn next_chunk_back<const N: usize>(
&mut self,
) -> Result<[Self::Item; N], IntoIter<Self::Item, N>>where
Self: Sized,
fn next_chunk_back<const N: usize>(
&mut self,
) -> Result<[Self::Item; N], IntoIter<Self::Item, N>>where
Self: Sized,
iter_next_chunk)N values in sequence. Read moreSourcefn advance_back_by(&mut self, n: usize) -> Result<(), NonZero<usize>>
fn advance_back_by(&mut self, n: usize) -> Result<(), NonZero<usize>>
iter_advance_by)n elements. Read more1.37.0 (const: unstable) · Sourcefn nth_back(&mut self, n: usize) -> Option<Self::Item>
fn nth_back(&mut self, n: usize) -> Option<Self::Item>
nth element from the end of the iterator. Read more1.27.0 (const: unstable) · Sourcefn try_rfold<B, F, R>(&mut self, init: B, f: F) -> R
fn try_rfold<B, F, R>(&mut self, init: B, f: F) -> R
Iterator::try_fold(): it takes
elements starting from the back of the iterator. Read moreimpl Eq for TimeSeries
impl ExactSizeIterator for TimeSerieswhere
TimeSeries: Iterator,
impl ExactSizeIterator for TimeSerieswhere
TimeSeries: Iterator,
impl<'py> IntoPyObject<'py> for TimeSeries
impl<'py> IntoPyObject<'py> for TimeSeries
type Target = TimeSeries
type Target = TimeSeries
type Output = Bound<'py, <TimeSeries as IntoPyObject<'py>>::Target>
type Output = Bound<'py, <TimeSeries as IntoPyObject<'py>>::Target>
fn into_pyobject(
self,
py: Python<'py>,
) -> Result<<TimeSeries as IntoPyObject<'py>>::Output, <TimeSeries as IntoPyObject<'py>>::Error>
fn into_pyobject( self, py: Python<'py>, ) -> Result<<TimeSeries as IntoPyObject<'py>>::Output, <TimeSeries as IntoPyObject<'py>>::Error>
impl Iterator for TimeSeries
impl Iterator for TimeSeries
fn size_hint(&self) -> (usize, Option<usize>)
fn size_hint(&self) -> (usize, Option<usize>)
Sourcefn next_chunk<const N: usize>(
&mut self,
) -> Result<[Self::Item; N], IntoIter<Self::Item, N>>where
Self: Sized,
fn next_chunk<const N: usize>(
&mut self,
) -> Result<[Self::Item; N], IntoIter<Self::Item, N>>where
Self: Sized,
iter_next_chunk)N values. Read more1.0.0 (const: unstable) · Sourcefn count(self) -> usizewhere
Self: Sized,
fn count(self) -> usizewhere
Self: Sized,
1.0.0 (const: unstable) · Sourcefn last(self) -> Option<Self::Item>where
Self: Sized,
fn last(self) -> Option<Self::Item>where
Self: Sized,
Sourcefn advance_by(&mut self, n: usize) -> Result<(), NonZero<usize>>
fn advance_by(&mut self, n: usize) -> Result<(), NonZero<usize>>
iter_advance_by)n elements. Read more1.0.0 (const: unstable) · Sourcefn nth(&mut self, n: usize) -> Option<Self::Item>
fn nth(&mut self, n: usize) -> Option<Self::Item>
nth element of the iterator. Read more1.28.0 (const: unstable) · Sourcefn step_by(self, step: usize) -> StepBy<Self> ⓘwhere
Self: Sized,
fn step_by(self, step: usize) -> StepBy<Self> ⓘwhere
Self: Sized,
1.0.0 (const: unstable) · Sourcefn chain<U>(self, other: U) -> Chain<Self, <U as IntoIterator>::IntoIter> ⓘ
fn chain<U>(self, other: U) -> Chain<Self, <U as IntoIterator>::IntoIter> ⓘ
1.0.0 (const: unstable) · Sourcefn zip<U>(self, other: U) -> Zip<Self, <U as IntoIterator>::IntoIter> ⓘwhere
Self: Sized,
U: IntoIterator,
fn zip<U>(self, other: U) -> Zip<Self, <U as IntoIterator>::IntoIter> ⓘwhere
Self: Sized,
U: IntoIterator,
Sourcefn intersperse(self, separator: Self::Item) -> Intersperse<Self> ⓘ
fn intersperse(self, separator: Self::Item) -> Intersperse<Self> ⓘ
iter_intersperse)separator between items
of the original iterator. Read moreSourcefn intersperse_with<G>(self, separator: G) -> IntersperseWith<Self, G> ⓘ
fn intersperse_with<G>(self, separator: G) -> IntersperseWith<Self, G> ⓘ
iter_intersperse)separator
between items of the original iterator. Read more1.0.0 (const: unstable) · Sourcefn map<B, F>(self, f: F) -> Map<Self, F> ⓘ
fn map<B, F>(self, f: F) -> Map<Self, F> ⓘ
1.21.0 (const: unstable) · Sourcefn for_each<F>(self, f: F)
fn for_each<F>(self, f: F)
1.0.0 (const: unstable) · Sourcefn filter<P>(self, predicate: P) -> Filter<Self, P> ⓘ
fn filter<P>(self, predicate: P) -> Filter<Self, P> ⓘ
1.0.0 (const: unstable) · Sourcefn filter_map<B, F>(self, f: F) -> FilterMap<Self, F> ⓘ
fn filter_map<B, F>(self, f: F) -> FilterMap<Self, F> ⓘ
1.0.0 (const: unstable) · Sourcefn enumerate(self) -> Enumerate<Self> ⓘwhere
Self: Sized,
fn enumerate(self) -> Enumerate<Self> ⓘwhere
Self: Sized,
1.0.0 (const: unstable) · Sourcefn take_while<P>(self, predicate: P) -> TakeWhile<Self, P> ⓘ
fn take_while<P>(self, predicate: P) -> TakeWhile<Self, P> ⓘ
1.57.0 (const: unstable) · Sourcefn map_while<B, P>(self, predicate: P) -> MapWhile<Self, P> ⓘ
fn map_while<B, P>(self, predicate: P) -> MapWhile<Self, P> ⓘ
1.0.0 (const: unstable) · Sourcefn skip(self, n: usize) -> Skip<Self> ⓘwhere
Self: Sized,
fn skip(self, n: usize) -> Skip<Self> ⓘwhere
Self: Sized,
n elements. Read more1.0.0 (const: unstable) · Sourcefn take(self, n: usize) -> Take<Self> ⓘwhere
Self: Sized,
fn take(self, n: usize) -> Take<Self> ⓘwhere
Self: Sized,
n elements, or fewer
if the underlying iterator ends sooner. Read more1.0.0 (const: unstable) · Sourcefn scan<St, B, F>(self, initial_state: St, f: F) -> Scan<Self, St, F> ⓘ
fn scan<St, B, F>(self, initial_state: St, f: F) -> Scan<Self, St, F> ⓘ
1.0.0 (const: unstable) · Sourcefn flat_map<U, F>(self, f: F) -> FlatMap<Self, U, F> ⓘ
fn flat_map<U, F>(self, f: F) -> FlatMap<Self, U, F> ⓘ
1.29.0 (const: unstable) · Sourcefn flatten(self) -> Flatten<Self> ⓘ
fn flatten(self) -> Flatten<Self> ⓘ
Sourcefn map_windows<F, R, const N: usize>(self, f: F) -> MapWindows<Self, F, N> ⓘ
fn map_windows<F, R, const N: usize>(self, f: F) -> MapWindows<Self, F, N> ⓘ
iter_map_windows)f for each contiguous window of size N over
self and returns an iterator over the outputs of f. Like slice::windows(),
the windows during mapping overlap as well. Read more1.0.0 (const: unstable) · Sourcefn inspect<F>(self, f: F) -> Inspect<Self, F> ⓘ
fn inspect<F>(self, f: F) -> Inspect<Self, F> ⓘ
1.0.0 (const: unstable) · Sourcefn by_ref(&mut self) -> &mut Selfwhere
Self: Sized,
fn by_ref(&mut self) -> &mut Selfwhere
Self: Sized,
Iterator. Read more1.0.0 (const: unstable) · Sourcefn collect<B>(self) -> B
fn collect<B>(self) -> B
Sourcefn try_collect<B>(
&mut self,
) -> <<Self::Item as Try>::Residual as Residual<B>>::TryType
fn try_collect<B>( &mut self, ) -> <<Self::Item as Try>::Residual as Residual<B>>::TryType
iterator_try_collect)Sourcefn collect_into<E>(self, collection: &mut E) -> &mut E
fn collect_into<E>(self, collection: &mut E) -> &mut E
iter_collect_into)1.0.0 (const: unstable) · Sourcefn partition<B, F>(self, f: F) -> (B, B)
fn partition<B, F>(self, f: F) -> (B, B)
Sourcefn partition_in_place<'a, T, P>(self, predicate: P) -> usize
fn partition_in_place<'a, T, P>(self, predicate: P) -> usize
iter_partition_in_place)true precede all those that return false.
Returns the number of true elements found. Read moreSourcefn is_partitioned<P>(self, predicate: P) -> bool
fn is_partitioned<P>(self, predicate: P) -> bool
iter_is_partitioned)true precede all those that return false. Read more1.27.0 (const: unstable) · Sourcefn try_fold<B, F, R>(&mut self, init: B, f: F) -> R
fn try_fold<B, F, R>(&mut self, init: B, f: F) -> R
1.27.0 (const: unstable) · Sourcefn try_for_each<F, R>(&mut self, f: F) -> R
fn try_for_each<F, R>(&mut self, f: F) -> R
1.0.0 (const: unstable) · Sourcefn fold<B, F>(self, init: B, f: F) -> B
fn fold<B, F>(self, init: B, f: F) -> B
1.51.0 (const: unstable) · Sourcefn reduce<F>(self, f: F) -> Option<Self::Item>
fn reduce<F>(self, f: F) -> Option<Self::Item>
Sourcefn try_reduce<R>(
&mut self,
f: impl FnMut(Self::Item, Self::Item) -> R,
) -> <<R as Try>::Residual as Residual<Option<<R as Try>::Output>>>::TryType
fn try_reduce<R>( &mut self, f: impl FnMut(Self::Item, Self::Item) -> R, ) -> <<R as Try>::Residual as Residual<Option<<R as Try>::Output>>>::TryType
iterator_try_reduce)1.0.0 (const: unstable) · Sourcefn all<F>(&mut self, f: F) -> bool
fn all<F>(&mut self, f: F) -> bool
1.0.0 (const: unstable) · Sourcefn any<F>(&mut self, f: F) -> bool
fn any<F>(&mut self, f: F) -> bool
1.0.0 (const: unstable) · Sourcefn find<P>(&mut self, predicate: P) -> Option<Self::Item>
fn find<P>(&mut self, predicate: P) -> Option<Self::Item>
1.30.0 (const: unstable) · Sourcefn find_map<B, F>(&mut self, f: F) -> Option<B>
fn find_map<B, F>(&mut self, f: F) -> Option<B>
Sourcefn try_find<R>(
&mut self,
f: impl FnMut(&Self::Item) -> R,
) -> <<R as Try>::Residual as Residual<Option<Self::Item>>>::TryType
fn try_find<R>( &mut self, f: impl FnMut(&Self::Item) -> R, ) -> <<R as Try>::Residual as Residual<Option<Self::Item>>>::TryType
try_find)1.0.0 (const: unstable) · Sourcefn position<P>(&mut self, predicate: P) -> Option<usize>
fn position<P>(&mut self, predicate: P) -> Option<usize>
1.0.0 (const: unstable) · Sourcefn rposition<P>(&mut self, predicate: P) -> Option<usize>
fn rposition<P>(&mut self, predicate: P) -> Option<usize>
1.0.0 (const: unstable) · Sourcefn max(self) -> Option<Self::Item>
fn max(self) -> Option<Self::Item>
1.0.0 (const: unstable) · Sourcefn min(self) -> Option<Self::Item>
fn min(self) -> Option<Self::Item>
1.6.0 (const: unstable) · Sourcefn max_by_key<B, F>(self, f: F) -> Option<Self::Item>
fn max_by_key<B, F>(self, f: F) -> Option<Self::Item>
1.15.0 (const: unstable) · Sourcefn max_by<F>(self, compare: F) -> Option<Self::Item>
fn max_by<F>(self, compare: F) -> Option<Self::Item>
1.6.0 (const: unstable) · Sourcefn min_by_key<B, F>(self, f: F) -> Option<Self::Item>
fn min_by_key<B, F>(self, f: F) -> Option<Self::Item>
1.15.0 (const: unstable) · Sourcefn min_by<F>(self, compare: F) -> Option<Self::Item>
fn min_by<F>(self, compare: F) -> Option<Self::Item>
1.0.0 (const: unstable) · Sourcefn rev(self) -> Rev<Self> ⓘwhere
Self: Sized + DoubleEndedIterator,
fn rev(self) -> Rev<Self> ⓘwhere
Self: Sized + DoubleEndedIterator,
1.0.0 (const: unstable) · Sourcefn unzip<A, B, FromA, FromB>(self) -> (FromA, FromB)
fn unzip<A, B, FromA, FromB>(self) -> (FromA, FromB)
1.36.0 (const: unstable) · Sourcefn copied<'a, T>(self) -> Copied<Self> ⓘ
fn copied<'a, T>(self) -> Copied<Self> ⓘ
1.0.0 (const: unstable) · Sourcefn cycle(self) -> Cycle<Self> ⓘ
fn cycle(self) -> Cycle<Self> ⓘ
Sourcefn array_chunks<const N: usize>(self) -> ArrayChunks<Self, N> ⓘwhere
Self: Sized,
fn array_chunks<const N: usize>(self) -> ArrayChunks<Self, N> ⓘwhere
Self: Sized,
iter_array_chunks)N elements of the iterator at a time. Read more1.11.0 (const: unstable) · Sourcefn product<P>(self) -> P
fn product<P>(self) -> P
Sourcefn cmp_by<I, F>(self, other: I, cmp: F) -> Ordering
fn cmp_by<I, F>(self, other: I, cmp: F) -> Ordering
iter_order_by)Iterator with those
of another with respect to the specified comparison function. Read more1.5.0 (const: unstable) · Sourcefn partial_cmp<I>(self, other: I) -> Option<Ordering>
fn partial_cmp<I>(self, other: I) -> Option<Ordering>
PartialOrd elements of
this Iterator with those of another. The comparison works like short-circuit
evaluation, returning a result without comparing the remaining elements.
As soon as an order can be determined, the evaluation stops and a result is returned. Read moreSourcefn partial_cmp_by<I, F>(self, other: I, partial_cmp: F) -> Option<Ordering>where
Self: Sized,
I: IntoIterator,
F: FnMut(Self::Item, <I as IntoIterator>::Item) -> Option<Ordering>,
fn partial_cmp_by<I, F>(self, other: I, partial_cmp: F) -> Option<Ordering>where
Self: Sized,
I: IntoIterator,
F: FnMut(Self::Item, <I as IntoIterator>::Item) -> Option<Ordering>,
iter_order_by)Iterator with those
of another with respect to the specified comparison function. Read moreSourcefn eq_by<I, F>(self, other: I, eq: F) -> bool
fn eq_by<I, F>(self, other: I, eq: F) -> bool
iter_order_by)1.5.0 (const: unstable) · Sourcefn lt<I>(self, other: I) -> bool
fn lt<I>(self, other: I) -> bool
Iterator are lexicographically
less than those of another. Read more1.5.0 (const: unstable) · Sourcefn le<I>(self, other: I) -> bool
fn le<I>(self, other: I) -> bool
Iterator are lexicographically
less or equal to those of another. Read more1.5.0 (const: unstable) · Sourcefn gt<I>(self, other: I) -> bool
fn gt<I>(self, other: I) -> bool
Iterator are lexicographically
greater than those of another. Read more1.5.0 (const: unstable) · Sourcefn ge<I>(self, other: I) -> bool
fn ge<I>(self, other: I) -> bool
Iterator are lexicographically
greater than or equal to those of another. Read more1.82.0 (const: unstable) · Sourcefn is_sorted(self) -> bool
fn is_sorted(self) -> bool
1.82.0 (const: unstable) · Sourcefn is_sorted_by<F>(self, compare: F) -> bool
fn is_sorted_by<F>(self, compare: F) -> bool
impl LowerExp for TimeSeries
impl LowerExp for TimeSeries
impl LowerHex for TimeSeries
impl LowerHex for TimeSeries
impl Octal for TimeSeries
impl Octal for TimeSeries
impl PartialEq for TimeSeries
impl PartialEq for TimeSeries
impl Pointer for TimeSeries
impl Pointer for TimeSeries
impl PyClass for TimeSeries
impl PyClass for TimeSeries
impl PyTypeInfo for TimeSeries
impl PyTypeInfo for TimeSeries
const NAME: &'static str = <Self as ::pyo3::PyClass>::NAME
const NAME: &'static str = <Self as ::pyo3::PyClass>::NAME
prefer using ::type_object(py).name() to get the correct runtime value
const MODULE: Option<&'static str> = <Self as ::pyo3::impl_::pyclass::PyClassImpl>::MODULE
const MODULE: Option<&'static str> = <Self as ::pyo3::impl_::pyclass::PyClassImpl>::MODULE
prefer using ::type_object(py).module() to get the correct runtime value
fn type_object_raw(py: Python<'_>) -> *mut PyTypeObject
fn type_object_raw(py: Python<'_>) -> *mut PyTypeObject
fn type_object(py: Python<'_>) -> Bound<'_, PyType>
fn type_object(py: Python<'_>) -> Bound<'_, PyType>
fn is_type_of(object: &Bound<'_, PyAny>) -> bool
fn is_type_of(object: &Bound<'_, PyAny>) -> bool
object is an instance of this type or a subclass of this type.fn is_exact_type_of(object: &Bound<'_, PyAny>) -> bool
fn is_exact_type_of(object: &Bound<'_, PyAny>) -> bool
object is an instance of this type.impl StructuralPartialEq for TimeSeries
impl UpperExp for TimeSeries
impl UpperExp for TimeSeries